
{% extends "layout.html" %}
{%- block htmltitle -%}
<title>{{ shorttitle }}</title>
{%- endblock -%}
{% block body %}
  <h1>{{ docstitle|e }}</h1>
  <p>
  Welcome! This is the documentation for neo-mamba.
  </p>
  <table class="contentstable" align="center"><tr>
    <td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("whatsnew/" + version) }}">What's new in neo-mamba {{ version }}?</a><br/>
        <span class="linkdescr"> or <a href="{{ pathto("whatsnew/index") }}">all "What's new" documents</a> since inception</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("tutorial/index") }}">Tutorial</a><br/>
         <span class="linkdescr">start here</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("library/index") }}">Library Reference</a><br/>
         <span class="linkdescr">keep this under your pillow</span></p>
    </td><td width="50%">
      <p class="biglink"><a class="biglink" href="{{ pathto("howto/index") }}">neo-mamba HOWTOs</a><br/>
         <span class="linkdescr">in-depth documents on specific topics</span></p>
      <p class="biglink"><a class="biglink" href="{{ pathto("faq/index") }}">FAQs</a><br/>
         <span class="linkdescr">frequently asked questions (with answers!)</span></p>
    </td></tr>
  </table>
  {% endblock %}